from typing import Optional, List, Any
import numpy as np
import cv2
import PIL

# IMPORT EAGERX
import eagerx
from eagerx import Space
import eagerx.core.register as register
from eagerx.utils.utils import Msg
from eagerx.core.entities import EngineNode
from eagerx.core.specs import NodeSpec, ObjectSpec


class CameraRender(EngineNode):
    @classmethod
    def make(
        cls,
        name: str,
        rate: float,
        process: int = eagerx.NEW_PROCESS,
        color: str = "cyan",
        shape: Optional[List[int]] = None,
        camera_idx: int = 0,
    ) -> NodeSpec:
        """CameraRender spec"""
        spec = cls.get_specification()

        # Modify default node params
        spec.config.update(name=name, rate=rate, process=process, color=color)
        spec.config.update(inputs=["tick"], outputs=["image"])

        # Modify custom node params
        spec.config.shape = shape if isinstance(shape, list) else [480, 480]
        spec.config.camera_idx = camera_idx

        # Set image space
        spec.outputs.image.space.update(low=0, high=255, shape=[spec.config.shape[0], spec.config.shape[1], 3])
        return spec

    def initialize(self, spec: NodeSpec, object_spec: ObjectSpec, simulator: Any):
        self.cam = None
        self.height, self.width = spec.config.shape
        self.camera_idx = spec.config.camera_idx
        self.render_toggle = False
        self.sub_toggle = self.backend.Subscriber("%s/env/render/toggle" % self.ns, "bool", self._set_render_toggle)

    @register.states()
    def reset(self):
        # This sensor is stateless (in contrast to e.g. a Kalman filter).
        pass

    @register.inputs(tick=Space(shape=(), dtype="int64"))  # Dummy, because a node must have at least one input.
    @register.outputs(image=Space(dtype="uint8"))
    def callback(self, t_n: float, tick: Optional[Msg] = None):
        if self.render_toggle:
            self.cam: cv2.VideoCapture
            ret, cv_img = self.cam.read()
            if ret:
                if cv_img.shape[:2] != (self.height, self.width):
                    img_PIL = PIL.Image.fromarray(cv_img).convert("RGB")
                    img_PIL = img_PIL.resize((self.height, self.width), PIL.Image.ANTIALIAS)
                    img = np.array(img_PIL).astype("uint8")
                else:
                    img = cv_img.astype("uint8")
            else:
                img = np.zeros((self.height, self.width, 3), dtype="uint8")
        else:
            img = np.zeros((self.height, self.width, 3), dtype="uint8")
        return dict(image=img)

    def _set_render_toggle(self, msg):
        if msg:
            self.backend.logdebug("[%s] START RENDERING!" % self.name)
            self.cam = cv2.VideoCapture(self.camera_idx)
            self.cam.set(cv2.CAP_PROP_FRAME_WIDTH, self.width)
            self.cam.set(cv2.CAP_PROP_FRAME_HEIGHT, self.height)
        else:
            self.backend.logdebug("[%s] STOPPED RENDERING!" % self.name)
            if self.cam is not None:
                self.cam.release()
                self.cam = None
        self.render_toggle = msg

    def shutdown(self):
        self.backend.logdebug(f"[{self.name}] {self.name}.shutdown() called.")
        self.sub_toggle.unregister()

        # Release camera resources
        if self.cam is not None:
            self.cam.release()
            self.cam = None



# OTHER
from typing import Optional

# RX IMPORTS
import eagerx
import eagerx.core.register as register
from eagerx.core.entities import Engine
from eagerx.core.specs import EngineSpec, ObjectSpec


class RealEngine(Engine):
    @classmethod
    def make(
        cls,
        rate,
        process: Optional[int] = eagerx.NEW_PROCESS,
        sync: Optional[bool] = False,
        log_level: Optional[int] = eagerx.ERROR,
    ) -> EngineSpec:
        """
        Spec of the RealEngine

        :param rate: Rate of the engine
        :param process: {0: NEW_PROCESS, 1: ENVIRONMENT, 2: ENGINE, 3: EXTERNAL}
        :param sync: A boolean flag to set the mode of synchronization.
        :param log_level: {0: SILENT, 10: DEBUG, 20: INFO, 30: WARN, 40: ERROR, 50: FATAL}
        :return: EngineSpec
        """
        spec = cls.get_specification()

        # Modify default engine params
        spec.config.update(rate=rate, process=process, sync=sync, real_time_factor=1, simulate_delays=False)
        spec.config.update(log_level=log_level, color="magenta")
        return spec

    def initialize(self, spec: EngineSpec):
        self.simulator = dict()

    def add_object(self, spec: ObjectSpec):
        object_name = spec.config.name
        entity_id = spec.config.entity_id

        # add object to simulator (we have a ref to the simulator with self.simulator)
        self.backend.loginfo(f'Adding object "{object_name}" of type "{entity_id}" to the simulator.')

        # Extract relevant agnostic_params
        obj_name = object_name

        # Create new env, and add to simulator
        self.simulator[obj_name] = dict(state=None, input=None)

    def pre_reset(self):
        pass

    @register.states()
    def reset(self):
        pass

    def callback(self, t_n: float):
        pass
